import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class CrochetGrammar extends PApplet {

// 0 = slip knot
// 1 = close slip knot
// 2 = turn (right)
// 3 = turn (left)
// 4 = chain
// 5 = slip stitch
// 6 = single crochet 
// 7 = half double crochet
// 8 = double crochet
// 9 = triple crochet
Stitch[] coreSts = new Stitch[10];

int currHeightAboveRow = 0;
int currWidthProg = 0;
int maxWidth = 10;
int prevAnchor = -1;
int currAnchor = -1;

Stitch currStitch;
Stitch lastStitch;
ArrayList stitches = new ArrayList(); 
ArrayList clusters = new ArrayList();
ArrayList pieces = new ArrayList();
ArrayList fillerPieces = new ArrayList();
ArrayList rows = new ArrayList();

PFont masterFont;
PFont titleFont;

public void setup() {
  size(1000, 600);
  smooth();

  masterFont = loadFont("ArialMT-11.vlw");
  titleFont = loadFont("Arial-Black-24.vlw");

  coreSts[0] = new Stitch(-1, 0, "slip knot", 0, 0, false, -1);
  coreSts[0].initGraphics(0, 1, 5, 5, loadImage("SlipKnotStart.png"));
  coreSts[1] = new Stitch(-1, 1, "close slip knot", 0, 0, true, -1);
  coreSts[1].initGraphics(20, 1, 5, 5, loadImage("SlipKnotEnd.png"));
  coreSts[2] = new Stitch(-1, 2, "turn right", 0, 0, false, -1);
  coreSts[2].initGraphics(40, 1, 0, 0, loadImage("SlipKnotStart.png"));
  coreSts[3] = new Stitch(-1, 3, "turn left", 0, 0, false, -1);
  coreSts[3].initGraphics(60, 1, 0, 0, loadImage("SlipKnotEnd.png"));
  coreSts[4] = new Stitch(-1, 4, "chain", 1, 0, false, -1);
  coreSts[4].initGraphics(80, 1, 20, 10, loadImage("ChainStitch.png"));
  coreSts[5] = new Stitch(-1, 5, "slip stitch", 1, 0, true, -1);
  coreSts[5].initGraphics(100, 1, 20, 10, loadImage("SlipStitch.png"));
  coreSts[6] = new Stitch(-1, 6, "single crochet", 1, 1, true, -1);
  coreSts[6].initGraphics(120, 1, 20, 20, loadImage("SingleCrochet.png"));
  coreSts[7] = new Stitch(-1, 7, "half double crochet", 1, 2, true, -1);
  coreSts[7].initGraphics(140, 1, 20, 40, loadImage("HalfDoubleCrochet.png"));
  coreSts[8] = new Stitch(-1, 8, "double crochet", 1, 3, true, -1);
  coreSts[8].initGraphics(160, 1, 20, 60, loadImage("DoubleCrochet.png"));
  coreSts[9] = new Stitch(-1, 9, "triple crochet", 1, 4, true, -1);
  coreSts[9].initGraphics(180, 1, 20, 80, loadImage("TripleCrochet.png"));

  loadUI();
}

public void update() {
  updateUI();
}

public void draw() {
  update();
  background(255);
  testDraw();
  drawUI();
}

public void testDraw() {
  /*
  for (int i = 0; i < coreSts.length; i++) {
    if (coreSts[i].needsAnchor) fill(255);
    else fill(0);
    image(coreSts[i].img, coreSts[i].x + uiWidth, coreSts[i].y);
    //rect(coreSts[i].x + uiWidth, coreSts[i].y, coreSts[i].drawW, coreSts[i].drawH);
  }*/
  for (int j = 0; j < stitches.size(); j++) {
    Stitch st = (Stitch) stitches.get(j);
    if (st.needsAnchor) fill(255);
    else fill(0);
    st.drawStitch((50 * (j+1)) + uiWidth, 100);
  }

  int stwid = 50;
  for (int k = 0; k < clusters.size(); k++) {
    Cluster clust = (Cluster) clusters.get(k);
    clust.drawCluster(stwid + uiWidth, 200);
    stwid += clust.wid + 25;
  }
  stwid = 50;
  for (int i = 0; i < pieces.size(); i++) {
    // println("piece " + i);
    PatternPiece piece = (PatternPiece) pieces.get(i);

    for (int j = 0; j < piece.clusters.size(); j++) {
      //println("clusters " + j);
      Cluster clust = (Cluster) piece.clusters.get(j);
      clust.drawCluster(stwid + uiWidth, 300);
      stwid += clust.wid + 25;
    }
  }
  stwid = 50;
  for (int i = 0; i < fillerPieces.size(); i++) {
    // println("piece " + i);
    PatternPiece piece = (PatternPiece) fillerPieces.get(i);

    for (int j = 0; j < piece.clusters.size(); j++) {
      //println("clusters " + j);
      Cluster clust = (Cluster) piece.clusters.get(j);
      clust.drawCluster(stwid + uiWidth, 400);
      stwid += clust.wid + 25;
    }
  }

  stwid = 50;
  for (int l = 0; l < rows.size(); l++) {
    Row r = (Row) rows.get(l);
    for (int i = 0; i < r.pieces.size(); i++) {
      // println("piece " + i);
      PatternPiece piece = (PatternPiece) r.pieces.get(i);

      for (int j = 0; j < piece.clusters.size(); j++) {
        //println("clusters " + j);
        Cluster clust = (Cluster) piece.clusters.get(j);
        clust.drawCluster(stwid + uiWidth, 500);
        stwid += clust.wid;
      }
    }
  }
}

public void mousePressed() {
  if (!pressedUI()) {
    //generateRand();
    for (int j = 0; j < stitches.size(); j++) {
      Stitch st = (Stitch) stitches.get(j);
      print(st.name + ", ");
    }
    println();
    for (int i = 0; i < clusters.size(); i++) {
      Cluster clust = (Cluster) clusters.get(i);
      println(clust.contents);
    }
    println();
    for (int j = 0; j < pieces.size(); j++) {
      PatternPiece piece = (PatternPiece) pieces.get(j);
      println("piece piece: ");
      for (int i = 0; i < piece.clusters.size(); i++) {
        Cluster clust = (Cluster) piece.clusters.get(i);
        println(clust.contents);
      }
    }
    println();
  }
}

public void mouseReleased() {
  releaseUI();
}

public void generateRand() {
  stitches = new ArrayList(); 
  clusters = new ArrayList();
  int rand = ((int) random(5)) + 1;
  for (int i = 0; i < rand; i++) {
    int stRand = ((int) random(coreSts.length-4)) + 4;
    // Note: Check height restrictions here
    stitches.add(coreSts[stRand].copyStitch());
    if (clusters.size() == 0) { // Make a new cluster and move on
      Cluster clust = new Cluster(coreSts[stRand].copyStitch());
      clusters.add(clust);
    } 
    else { // Check to see if you want to add it to the previous cluster
      int clustRand = (int) random(2);
      if (clustRand == 0) {
        // make a new cluster
        Cluster clust = new Cluster(coreSts[stRand].copyStitch());
        clusters.add(clust);
      } 
      else {
        Cluster clust = (Cluster) clusters.get(clusters.size()-1);
        clust.appendStitch(coreSts[stRand].copyStitch());
      }
    }
  }
  PatternPiece piece = new PatternPiece(clusters);
  pieces.add(piece);
}

// Note: Check height restrictions... somewhere
public void generateStitches() {
  stitches = new ArrayList(); 

  // Determine how many stitches we will be generating
  int numToGenerate = 0;
  if (numStitches == 0) numToGenerate = ((int) random(5)) + 1;
  else numToGenerate = numStitches;

  // Generate the stitches to fill the num with chance of repetition
  for (int i = 0; i < numToGenerate; i++) {
    // Generate a random first stitch
    if (i == 0) {
      int stRand = ((int) random(coreSts.length-4)) + 4;
      stitches.add(coreSts[stRand].copyStitch());
    } 
    else {
      // If there is a repeat chance, check it
      int repeatRand = ((int) random(100));
      if (repeatRand < chanceOfRepetition) {
        // If the 0-100 random number < number selected, repeat the last stitch
        Stitch lastStitch = (Stitch) stitches.get(i-1);
        stitches.add(coreSts[lastStitch.coreStID].copyStitch());
      } 
      else { // Else, find a completely random stitch
        int stRand = ((int) random(coreSts.length-4)) + 4;
        stitches.add(coreSts[stRand].copyStitch());
      }
    }
  }

  // Now we have x number of stitches with repetition
  // Apply symmetry, if any, and if we do, check if we need to input a chain.
  int symmetryRand = ((int) random(100));
  if (symmetryRand < chanceOfSymmetry) {
    symmetrical = true;
    int middleID = numToGenerate/2;
    // If we have an odd number of stitches...
    if (insertChainAtPeak && numToGenerate % 2 == 1) {
      // If chain @ peak is selected, replace the middle stitch with a chain
      stitches.set(middleID, coreSts[4].copyStitch());
    } 
    else if (insertChainAtPeak) {
      // If chain @ peak is selected, insert a stitch in the middle
      stitches.add(middleID, coreSts[4].copyStitch());
    }
    // Now repeat the left side on the right side in mirrored form.
    for (int j = 0; j < middleID; j++) {
      Stitch st = (Stitch) stitches.get(j);
      int reflectionID = stitches.size()-1-j;
      stitches.set(reflectionID, st.copyStitch());
    }
  } 
  else {
    symmetrical = false;
    // Do nothing! No Symmetry, no Symmetrical peak, the creation of stitches is done.
  }
  generateClusters();
}

public void generateClusters() {
  clusters = new ArrayList();
  pieces = new ArrayList();

  // Determining if we're clustering
  int clusterRand = ((int) random(100));
  if (clusterRand < chanceToGroup) clustersOn = true;
  else clustersOn = false;

  if (clustersOn) {
    int numToCluster = ((int) random(6)) + 2; // Group 2-7 stitches
    if (symmetrical) {
      // Cluster around the symmetrical midpoint
      int middleID = stitches.size()/2;
      Cluster clust = null;
      // Odd number of stitches to cluster, make sure to include the peak
      int startIndex;
      if (stitches.size() % 2 == 0 && numToCluster % 2 == 0) startIndex = middleID - (numToCluster/2) -1;
      else startIndex = middleID - (numToCluster/2);
      for (int i = 0; i < stitches.size(); i++) {
        // any pre- or post-cluster, add them as they're own cluster
        if (i < startIndex || i > startIndex + numToCluster) {
          Stitch st = (Stitch) stitches.get(i);
          clust = new Cluster(st.copyStitch());
          clusters.add(clust);
          clust = null;
        } 
        else { // Else, add them to the working cluster
          // if the new cluster hasn't been started yet, add it
          if (clust == null) clust = new Cluster();
          Stitch st = (Stitch) stitches.get(i);
          clust.appendStitch(st.copyStitch());
        }
        if (i == startIndex + numToCluster) {
          // Finish off the cluster and do single from then on 
          clusters.add(clust);
          clust = null;
        }
      }
    } 
    else {
      Cluster clust = null;
      int clusterCt = 0;
      for (int i = 0; i < stitches.size(); i++) {
        if (clust == null) clust = new Cluster();
        Stitch st = (Stitch) stitches.get(i);
        clust.appendStitch(st.copyStitch());
        clusterCt++;
        if (i == stitches.size()-1 || clusterCt == numToCluster || clusterRand >= chanceToGroup) {
          clusters.add(clust);
          clust = null;
          clusterCt = 0;
        }
        clusterRand = ((int) random(100));
      }
    }
  } 
  else {
    // If clusters aren't turned on, dump them each into their own cluster and move on.
    for (int i = 0; i < stitches.size(); i++) {
      Stitch st = (Stitch) stitches.get(i);
      Cluster clust = new Cluster(st.copyStitch());
      clusters.add(clust);
    }
  }
  pieces.add(new PatternPiece(clusters));
  generateFillerStitches();
  generateAlternateCluster();
  generateRow();
}

public void generateFillerStitches() {
  fillerPieces = new ArrayList();
  // If filler stitches is turned on...
  if (numFillerStitches > 0) {
    // Pick a random stitch that requires an anchor
    int stRand = ((int) random(coreSts.length-5)) + 5;
    Stitch st = coreSts[stRand].copyStitch();
    PatternPiece p = new PatternPiece();
    // And fill up a cluster with the specified number of that stitch
    for (int i = 0; i < numFillerStitches; i++) {
      Cluster clust = new Cluster(st.copyStitch());
      p.appendCluster(clust);
    }
    // Keep track of this piece separately for later use
    fillerPieces.add(p);
  }
}

public void generateAlternateCluster() {
  PatternPiece reversePiece = new PatternPiece();
  if (horizontalReversal && pieces.size() > 0) {
    PatternPiece p = (PatternPiece) pieces.get(0);
    // For each cluster, starting with the last one
    for (int i = p.clusters.size() -1; i >= 0; i--) {
      Cluster clustOld = (Cluster) p.clusters.get(i);
      Cluster clustNew = new Cluster();
      // Reverse the stitches found in it!
      for (int j = clustOld.clustSts.size()-1; j >= 0; j--) {
        Stitch st = (Stitch)clustOld.clustSts.get(j);
        clustNew.appendStitch(st.copyStitch());
      }
      reversePiece.appendCluster(clustNew);
    }
    pieces.add(reversePiece);
  }
}

public void generateRow() {
  rows = new ArrayList();
  Row r = new Row();
  PatternPiece p;
  int lengthOfRow = lengthRows;

  // Start with filler, end with filler
  if (numFillerStitches > 0) {
    for (int k = 0; k < fillerPieces.size(); k++) {
      p = (PatternPiece) fillerPieces.get(k);
      r.appendPiece(p);
    }
  }
  lengthOfRow--;

  // Fill in the rest of the pattern
  for (int i = 0; i < lengthOfRow; i++) {
    for (int j = 0; j < pieces.size(); j++) {
      p = (PatternPiece) pieces.get(j);
      r.appendPiece(p);
    }
    for (int k = 0; k < fillerPieces.size(); k++) {
      p = (PatternPiece) fillerPieces.get(k);
      r.appendPiece(p);
    }
  }
  rows.add(r);
}

class Button {
  int id;
  int x;
  int y;
  int w;
  int h;
  String words;
  PFont font;

  int hover;
  int click;
  int standard;

  Boolean hoverCheck = false;
  Boolean clickCheck = false;
  Boolean clicked = false;

  Button(int id, int x, int y, int w, int h, String words, PFont font ) {
    this.id = id;
    this.x = x;
    this.y = y;
    this.w = w;
    this.h = h;
    this.words = words;
    this.font = font;

    standard = color(250, 250, 250);
    hover = color(225, 225, 225);
    click = color(150, 150, 150);
  }

  public void drawButton() {
    stroke(0);
    strokeWeight(2);
    if (clickCheck) {
      fill(click);
    } 
    else if (hoverCheck) {
      fill (hover);
    } 
    else {
      fill(standard);
    }
    rect(x, y, w, h);
    textFont(font);
    fill(0);
    text(words, x+5, y+(h/2) + 5);
  }

  public void updateButton() {
    // check mouseover
    if (mouseX > x && mouseX < (x+w) && mouseY > y && mouseY < (y+h)) hoverCheck = true;
    else hoverCheck = false;
  }

  public boolean checkCollision() {
    if (hoverCheck) return true;
    else return false;
  }

  public void collide() {
    //println("** Button clicked: " + words + " **");
    clickCheck = true;
  }
}

class CheckBox{
  int id;
  int x;
  int y;
  int s;

  int hover;
  int click;
  int standard;
  
  Boolean hoverCheck = false;
  Boolean clickCheck = false;
  Boolean clicked = false;
  
  CheckBox(int id, int x, int y, int s) {
    this.id = id;
    this.x = x;
    this.y = y;
    this.s = s;
    
    standard = color(250, 250, 250);
    hover = color(225, 225, 225);
    click = color(150, 150, 150);
    
  }
  
  public void drawCheckBox(){
    stroke(0);
    strokeWeight(2);
    
    if (clickCheck){
      fill(click);
    } else if (hoverCheck) {
      fill (hover);
    } else {
      fill(standard);
    }
    rect(x, y, s, s);
    
    if (clicked){
      fill(click);
      rect(x+3, y+3, s-6, s-6);
    }
  }
  
  public void updateCheckBox(){
    // check mouseover
    if (mouseX > x && mouseX < (x+s) && mouseY > y && mouseY < (y+s)) hoverCheck = true;
    else hoverCheck = false;
  }
  
  public boolean checkCollision(){
    if (hoverCheck) return true;
    else return false;
  }
  
  public void collide(){
    //println("** Button clicked: " + words + " **");
    clickCheck = true;
  }
  
  public void release(){
    if (clickCheck){
      if(clicked == true) clicked = false;
      else clicked = true;
      clickCheck = false;
    }
  }
}
class Cluster{
  String contents = "";
  ArrayList clustSts = new ArrayList(); 
  int drawW;
  
  int wid = 0;
  int numStitches = 0;
  
  Cluster(ArrayList sts){
    clustSts = sts;
    numStitches = sts.size();
  }
  
  Cluster(Stitch st){
    appendStitch(st);
  }
  
  Cluster(){
  }
  
  public void drawCluster(int x, int y){
    int stwid = x;
    for(int i = 0; i < clustSts.size(); i++){
      Stitch st = (Stitch) clustSts.get(i);
      st.drawStitch(stwid, y);
      stwid += st.drawW;
    }
    wid = stwid-x;
  }
  
  public void appendStitch(Stitch st){
    drawW += st.drawW;
    contents += st.name + ";";
    numStitches++;
    clustSts.add(st);
  }
}
class Label{
  int id;
  int x;
  int y;
  String words;
  PFont font;

  Label(int id, int x, int y, String words, PFont font ) {
    this.id = id;
    this.x = x;
    this.y = y;
    this.words = words;
    this.font = font;
  }
  
  public void drawLabel(){
    textFont(font);
    fill(0);
    stroke(0);
    strokeWeight(1);
    text(words, x, y);
  }
}
class PatternPiece{
  ArrayList clusters = new ArrayList(); 
  int numStitches = 0;
  int numClusters = 0;
  
  PatternPiece(ArrayList clusts) {
    clusters = clusts;
    for (int i = 0; i < clusts.size(); i++){
      Cluster c = (Cluster) clusts.get(i);
      numStitches += c.numStitches;
      numClusters++;
    }
  }
  
  PatternPiece(){
  }
  
  public void appendCluster(Cluster clust){
    numStitches += clust.numStitches;
    numClusters++;
    clusters.add(clust);
  }
}
class Row{
  ArrayList pieces = new ArrayList(); 
  int numStitches = 0;
  int numClusters = 0;
  int numPieces = 0;
  
  Row(ArrayList piecess) {
    pieces = piecess;
    for (int i = 0; i < piecess.size(); i++){
      PatternPiece p = (PatternPiece) piecess.get(i);
      numStitches += p.numStitches;
      numClusters += p.numClusters;
      numPieces++;
    }
  }
  
  Row(){
  }
  
  public void appendPiece(PatternPiece p){
    numStitches += p.numStitches;
    numClusters += p.numClusters;
    numPieces++;
    pieces.add(p);
  }
}
class Slider{
  int id;
  int x;
  int y;
  int w;
  int h;
  PFont font;
  
  int value = 0; // numerical value 0-100
  int sliderPoint = 0; // the x value of the line we are pointing to
  
  Boolean hoverCheck = false;
  Boolean clickCheck = false;
  Boolean percentage = false;
  
  Slider(int id, int x, int y, int w, int h, Boolean percent, PFont font){
    this.id = id;
    this.x = x; 
    this.y = y;
    this.w = w;
    this.h = h;
    this.font = font;
    percentage = percent;
    
    value = 0;
    sliderPoint = x + (value * (w/10));
  }
  
  // MAGIC NUMBERS WHEEEEE
  public void drawSlider(){
    stroke(0);
    strokeWeight(1);
    line(x, y+h-2, x+w, y+h-2); // long bar
    for (int i = x; i <(x+w); i+=(w/10)){
      line(i, y+h-1, i, y+h-3); // dashes every 10 values out of 100
    }
    line(x, y+h-5, x, y+h); // Front marker
    line(x+w, y+h-5, x+w, y+h); // End marker
    
    if (hoverCheck) fill(0);
    else fill(255);
    triangle(sliderPoint, y+h-4, sliderPoint + (h/4), y, sliderPoint - (h/4), y);
    
    textFont(font);
    fill(0);
    if (percentage) text(value + " %", x+w + 5, y+(h/2) + 5);
    else text(value, x+w + 5, y+(h/2) + 5);
  }
  
  public void updateSlider(){
    int sliderX = sliderPoint - (h/4);
    int sliderY = y;
    int sliderW = h/2;
    int sliderH = h-4;
    
    // check mouseover
    if ((mouseX > sliderX && mouseX < (sliderX + sliderW) && 
        mouseY > sliderY && mouseY < (sliderY + sliderH)) ||
        (mouseX > x && mouseX < (x+w) && mouseY > (y+h-5) && mouseY < (y+h))) 
        hoverCheck = true;
    else hoverCheck = false;
    
    if(clickCheck){
      if (mouseX <= x) sliderPoint = x;
      else if (mouseX >= x + w) sliderPoint = x+w;
      else sliderPoint = mouseX;
      
      value = ((sliderPoint-x)/(w/100)); // Point / step, where step is w/100 (ie 1/100)
    }
  }
  
  public boolean checkCollision(){
    if (hoverCheck) return true;
    else return false;
  }
  
  public void collide(){
    //println("** Slider clicked: " + value + " **");
    clickCheck = true;
  }
  
  public void release(){
    //println("-- Slider released: " + value + " --");
    clickCheck = false;
  }
}
class Stitch{
  // unique ID
  int ID;
  // coreStitch ID
  int coreStID;
  // slst, ch, sc, etc
  String name;
  // 0-1, determines if we move on to the next anchor
  int w;
  // 0-5, keeps track of stitch heights
  int h;
  // Has a "hook in" as part of its makeup
  Boolean needsAnchor;
  // If it needs an achor, which ID is it?
  int anchorStitch;
  // The image file of the stitch
  PImage img;
  
  // Graphics
  int x;
  int y;
  int drawW;
  int drawH;
  
  Stitch(int id, int coreID, String n, int wid, int hei, Boolean na, int as){
    ID = id;
    coreStID = coreID;
    name = n;
    w = wid;
    h = hei;
    needsAnchor = na;
    anchorStitch = as;
  }
  
  public void initGraphics(int x, int y, int wid, int hei, PImage img){
    this.x = x;
    this.y = y;
    drawW = wid;
    drawH = hei;
    this.img = img;
  }
  
  public void drawStitch(int x, int y){
    this.x = x;
    this.y = y;
    image(img, x, y - drawH);
  }
  
  public Stitch copyStitch(){
    Stitch st = new Stitch(ID, coreStID, name, w, h, needsAnchor, anchorStitch);
    st.initGraphics(x, y, drawW, drawH, img);
    return st;
  }
}
// Data structures to hold all the UI elements
ArrayList titleLabels = new ArrayList();
ArrayList titleButtons = new ArrayList();

ArrayList genSliders = new ArrayList();
ArrayList genCheckboxes = new ArrayList();
ArrayList genLabels = new ArrayList();

ArrayList holdCheckboxes = new ArrayList();
ArrayList holdLabels = new ArrayList();

ArrayList finishSliders = new ArrayList();
ArrayList finishLabels = new ArrayList();

// Variables controlled by the UI
int numStitches = 0;
int chanceOfRepetition = 0;
int chanceOfSymmetry = 0;
Boolean insertChainAtPeak = false;
int chanceToGroup = 0;
Boolean groupAroundChains = false;
int numFillerStitches = 0;
Boolean horizontalReversal = false;
int amountOffsetRows = 0;
int numFillerRows = 0;
int lengthRows = 5; // ADD UI for me

// Characteristics of generated output
Boolean symmetrical = false;
Boolean clustersOn = false;

// Drawing info
int uiWidth = 250;

// Mode of UI: generate (0), hold (1), or finish (2)
int MODE = 0;

public void loadUI(){
  titleLabels.add(new Label(0, 30, 30, "Control Panel", titleFont));
  
  genLabels.add(new Label(1, 7, 65, "# of Stitches", masterFont));
  genSliders.add(new Slider(2, 12, 70, 200, 13, false, masterFont));
  genLabels.add(new Label(3, 7, 105, "% Chance of Repetition", masterFont));
  genSliders.add(new Slider(4, 12, 110, 200, 13, true, masterFont));
  genLabels.add(new Label(5, 7, 145, "% Chance of Symmetry", masterFont));
  genSliders.add(new Slider(6, 12, 150, 200, 13, true, masterFont));
  genLabels.add(new Label(7, 7, 190, "Insert Chain at Symmetrical Peak", masterFont));
  genCheckboxes.add(new CheckBox(8, 177, 180, 12));
  
  genLabels.add(new Label(9, 7, 250, "% Chance to Group into one Anchor", masterFont));
  genSliders.add(new Slider(10, 12, 255, 200, 13, true, masterFont));
  genLabels.add(new Label(11, 7, 295, "Group Around Chains, if any", masterFont));
  genCheckboxes.add(new CheckBox(12, 157, 285, 12));
  
  genLabels.add(new Label(13, 7, 350, "# of Filler Stitches (Between Clusters)", masterFont));
  genSliders.add(new Slider(14, 12, 355, 200, 13, false, masterFont));
  genLabels.add(new Label(15, 7, 395, "Alternate Horizontal Reversal (if applicable)", masterFont));
  genCheckboxes.add(new CheckBox(16, 225, 385, 12));
  
  genLabels.add(new Label(17, 7, 455, "% Offset Even Rows", masterFont));
  genSliders.add(new Slider(18, 12, 460, 200, 13, true, masterFont));
  genLabels.add(new Label(19, 7, 495, "# of Filler Rows", masterFont));
  genSliders.add(new Slider(20, 12, 500, 200, 13, true, masterFont));
  
  titleButtons.add(new Button(21, 90, height-50, 60, 30, "Generate", masterFont));
}

public void updateUI(){
  for(int i = 0; i < titleButtons.size(); i++){
    Button b = (Button) titleButtons.get(i);
    b.updateButton();
  }
  for(int i = 0; i < genSliders.size(); i++){
    Slider s = (Slider) genSliders.get(i);
    s.updateSlider();
  }
  for(int i = 0; i < genCheckboxes.size(); i++){
    CheckBox cb = (CheckBox) genCheckboxes.get(i);
    cb.updateCheckBox();
  }
  for(int i = 0; i < holdCheckboxes.size(); i++){
    CheckBox cb = (CheckBox) holdCheckboxes.get(i);
    cb.updateCheckBox();
  }
  for(int i = 0; i < finishSliders.size(); i++){
    Slider s = (Slider) finishSliders.get(i);
    s.updateSlider();
  }
}

public void drawUI(){
  stroke(0);
  strokeWeight(1);
  fill(230);
  rect(0, 0, uiWidth, height);
  
  for(int i = 0; i < titleLabels.size(); i++){
    Label l = (Label) titleLabels.get(i);
    l.drawLabel();
  }
  for(int i = 0; i < titleButtons.size(); i++){
    Button b = (Button) titleButtons.get(i);
    b.drawButton();
  }
  for(int i = 0; i < genSliders.size(); i++){
    Slider s = (Slider) genSliders.get(i);
    s.drawSlider();
  }
  for(int i = 0; i < genCheckboxes.size(); i++){
    CheckBox cb = (CheckBox) genCheckboxes.get(i);
    cb.drawCheckBox();
  }
  for(int i = 0; i < genLabels.size(); i++){
    Label l = (Label) genLabels.get(i);
    l.drawLabel();
  }
  for(int i = 0; i < holdCheckboxes.size(); i++){
    CheckBox cb = (CheckBox) holdCheckboxes.get(i);
    cb.drawCheckBox();
  }
  for(int i = 0; i < holdLabels.size(); i++){
    Label l = (Label) holdLabels.get(i);
    l.drawLabel();
  }
  for(int i = 0; i < finishSliders.size(); i++){
    Slider s = (Slider) finishSliders.get(i);
    s.drawSlider();
  }
  for(int i = 0; i < finishLabels.size(); i++){
    Label l = (Label) finishLabels.get(i);
    l.drawLabel();
  }
}

public Boolean pressedUI(){
  Boolean collisionDetected = false;
  
  if (mouseX > 0 && mouseX < uiWidth && mouseY > 0 && mouseY < height) collisionDetected = true;
  
  for(int i = 0; i < titleButtons.size(); i++){
    Button b = (Button) titleButtons.get(i);
    if (b.checkCollision()){
      collisionDetected = true;
      b.collide();
    }
  }
  
  for(int i = 0; i < genSliders.size(); i++){
    Slider s = (Slider) genSliders.get(i);
    if (s.checkCollision()){
      collisionDetected = true;
      s.collide();
    }
  }
  for(int i = 0; i < genCheckboxes.size(); i++){
    CheckBox cb = (CheckBox) genCheckboxes.get(i);
    if (cb.checkCollision()){
      collisionDetected = true;
      cb.collide();
    }
  }
  for(int i = 0; i < holdCheckboxes.size(); i++){
    CheckBox cb = (CheckBox) holdCheckboxes.get(i);
    if (cb.checkCollision()){
      collisionDetected = true;
      cb.collide();
    }
  }
  for(int i = 0; i < finishSliders.size(); i++){
    Slider s = (Slider) finishSliders.get(i);
    if (s.checkCollision()){
      collisionDetected = true;
      s.collide();
    }
  }
  
  // Returns true if any UI element is collided with
  return collisionDetected;
}

public void releaseUI(){
  for(int i = 0; i < titleButtons.size(); i++){
    Button b = (Button) titleButtons.get(i);
    if (b.clickCheck){ 
      //println("!!! RESOLVE BUTTON CLICK HERE !!!");
      if(b.id == 21) generateStitches();
      b.clickCheck = false;
    }
  }
  for(int i = 0; i < genSliders.size(); i++){
    Slider s = (Slider) genSliders.get(i);
    s.release();
    switch(s.id){
      case 2:
        numStitches = s.value;
        break;
      case 4:
        chanceOfRepetition = s.value;
        break;
      case 6:
        chanceOfSymmetry = s.value;
        break;
      case 10:
        chanceToGroup = s.value;
        break;
      case 14:
        numFillerStitches = s.value;
        break;
      case 18:
        amountOffsetRows = s.value;
        break;
      case 20:
        numFillerRows = s.value;
        break;
      default:
        break;
    }
  }
  for(int i = 0; i < genCheckboxes.size(); i++){
    CheckBox cb = (CheckBox) genCheckboxes.get(i);
    cb.release();
    switch(cb.id){
      case 8:
        if(cb.clicked) insertChainAtPeak = true;
        else insertChainAtPeak = false;
        break;
      case 12:
        if (cb.clicked) groupAroundChains = true;
        else groupAroundChains = false;
        break;
      case 16:
        if (cb.clicked) horizontalReversal = true;
        else horizontalReversal = false;
        break;
      default:
        break;
    }
  }
  
  // snapshot of variables
  /*
  println("--- UI Variables ---");
  println(numStitches);
  println(chanceOfRepetition);
  println(chanceOfSymmetry);
  println(insertChainAtPeak);
  println(chanceToGroup);
  println(groupAroundChains);
  println(numFillerStitches);
  println(horizontalReversal);
  println(amountOffsetRows);
  println(numFillerRows);
  println("----------------------");
  */
}

  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "CrochetGrammar" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
