// 0 = slip knot
// 1 = close slip knot
// 2 = turn (right)
// 3 = turn (left)
// 4 = chain
// 5 = slip stitch
// 6 = single crochet 
// 7 = half double crochet
// 8 = double crochet
// 9 = triple crochet
Stitch[] coreSts = new Stitch[10];

int currHeightAboveRow = 0;
int currWidthProg = 0;
int maxWidth = 10;
int prevAnchor = -1;
int currAnchor = -1;

Stitch currStitch;
Stitch lastStitch;
ArrayList stitches = new ArrayList(); 
ArrayList clusters = new ArrayList();
ArrayList pieces = new ArrayList();
ArrayList fillerPieces = new ArrayList();
ArrayList rows = new ArrayList();

PFont masterFont;
PFont titleFont;

void setup() {
  size(1000, 600);
  smooth();

  masterFont = loadFont("ArialMT-11.vlw");
  titleFont = loadFont("Arial-Black-24.vlw");

  coreSts[0] = new Stitch(-1, 0, "slip knot", 0, 0, false, -1);
  coreSts[0].initGraphics(0, 1, 5, 5, loadImage("SlipKnotStart.png"));
  coreSts[1] = new Stitch(-1, 1, "close slip knot", 0, 0, true, -1);
  coreSts[1].initGraphics(20, 1, 5, 5, loadImage("SlipKnotEnd.png"));
  coreSts[2] = new Stitch(-1, 2, "turn right", 0, 0, false, -1);
  coreSts[2].initGraphics(40, 1, 0, 0, loadImage("SlipKnotStart.png"));
  coreSts[3] = new Stitch(-1, 3, "turn left", 0, 0, false, -1);
  coreSts[3].initGraphics(60, 1, 0, 0, loadImage("SlipKnotEnd.png"));
  coreSts[4] = new Stitch(-1, 4, "chain", 1, 0, false, -1);
  coreSts[4].initGraphics(80, 1, 20, 10, loadImage("ChainStitch.png"));
  coreSts[5] = new Stitch(-1, 5, "slip stitch", 1, 0, true, -1);
  coreSts[5].initGraphics(100, 1, 20, 10, loadImage("SlipStitch.png"));
  coreSts[6] = new Stitch(-1, 6, "single crochet", 1, 1, true, -1);
  coreSts[6].initGraphics(120, 1, 20, 20, loadImage("SingleCrochet.png"));
  coreSts[7] = new Stitch(-1, 7, "half double crochet", 1, 2, true, -1);
  coreSts[7].initGraphics(140, 1, 20, 40, loadImage("HalfDoubleCrochet.png"));
  coreSts[8] = new Stitch(-1, 8, "double crochet", 1, 3, true, -1);
  coreSts[8].initGraphics(160, 1, 20, 60, loadImage("DoubleCrochet.png"));
  coreSts[9] = new Stitch(-1, 9, "triple crochet", 1, 4, true, -1);
  coreSts[9].initGraphics(180, 1, 20, 80, loadImage("TripleCrochet.png"));

  loadUI();
}

void update() {
  updateUI();
}

void draw() {
  update();
  background(255);
  testDraw();
  drawUI();
}

void testDraw() {
  /*
  for (int i = 0; i < coreSts.length; i++) {
    if (coreSts[i].needsAnchor) fill(255);
    else fill(0);
    image(coreSts[i].img, coreSts[i].x + uiWidth, coreSts[i].y);
    //rect(coreSts[i].x + uiWidth, coreSts[i].y, coreSts[i].drawW, coreSts[i].drawH);
  }*/
  for (int j = 0; j < stitches.size(); j++) {
    Stitch st = (Stitch) stitches.get(j);
    if (st.needsAnchor) fill(255);
    else fill(0);
    st.drawStitch((50 * (j+1)) + uiWidth, 100);
  }

  int stwid = 50;
  for (int k = 0; k < clusters.size(); k++) {
    Cluster clust = (Cluster) clusters.get(k);
    clust.drawCluster(stwid + uiWidth, 200);
    stwid += clust.wid + 25;
  }
  stwid = 50;
  for (int i = 0; i < pieces.size(); i++) {
    // println("piece " + i);
    PatternPiece piece = (PatternPiece) pieces.get(i);

    for (int j = 0; j < piece.clusters.size(); j++) {
      //println("clusters " + j);
      Cluster clust = (Cluster) piece.clusters.get(j);
      clust.drawCluster(stwid + uiWidth, 300);
      stwid += clust.wid + 25;
    }
  }
  stwid = 50;
  for (int i = 0; i < fillerPieces.size(); i++) {
    // println("piece " + i);
    PatternPiece piece = (PatternPiece) fillerPieces.get(i);

    for (int j = 0; j < piece.clusters.size(); j++) {
      //println("clusters " + j);
      Cluster clust = (Cluster) piece.clusters.get(j);
      clust.drawCluster(stwid + uiWidth, 400);
      stwid += clust.wid + 25;
    }
  }

  stwid = 50;
  for (int l = 0; l < rows.size(); l++) {
    Row r = (Row) rows.get(l);
    for (int i = 0; i < r.pieces.size(); i++) {
      // println("piece " + i);
      PatternPiece piece = (PatternPiece) r.pieces.get(i);

      for (int j = 0; j < piece.clusters.size(); j++) {
        //println("clusters " + j);
        Cluster clust = (Cluster) piece.clusters.get(j);
        clust.drawCluster(stwid + uiWidth, 500);
        stwid += clust.wid;
      }
    }
  }
}

void mousePressed() {
  if (!pressedUI()) {
    //generateRand();
    for (int j = 0; j < stitches.size(); j++) {
      Stitch st = (Stitch) stitches.get(j);
      print(st.name + ", ");
    }
    println();
    for (int i = 0; i < clusters.size(); i++) {
      Cluster clust = (Cluster) clusters.get(i);
      println(clust.contents);
    }
    println();
    for (int j = 0; j < pieces.size(); j++) {
      PatternPiece piece = (PatternPiece) pieces.get(j);
      println("piece piece: ");
      for (int i = 0; i < piece.clusters.size(); i++) {
        Cluster clust = (Cluster) piece.clusters.get(i);
        println(clust.contents);
      }
    }
    println();
  }
}

void mouseReleased() {
  releaseUI();
}

void generateRand() {
  stitches = new ArrayList(); 
  clusters = new ArrayList();
  int rand = ((int) random(5)) + 1;
  for (int i = 0; i < rand; i++) {
    int stRand = ((int) random(coreSts.length-4)) + 4;
    // Note: Check height restrictions here
    stitches.add(coreSts[stRand].copyStitch());
    if (clusters.size() == 0) { // Make a new cluster and move on
      Cluster clust = new Cluster(coreSts[stRand].copyStitch());
      clusters.add(clust);
    } 
    else { // Check to see if you want to add it to the previous cluster
      int clustRand = (int) random(2);
      if (clustRand == 0) {
        // make a new cluster
        Cluster clust = new Cluster(coreSts[stRand].copyStitch());
        clusters.add(clust);
      } 
      else {
        Cluster clust = (Cluster) clusters.get(clusters.size()-1);
        clust.appendStitch(coreSts[stRand].copyStitch());
      }
    }
  }
  PatternPiece piece = new PatternPiece(clusters);
  pieces.add(piece);
}

// Note: Check height restrictions... somewhere
void generateStitches() {
  stitches = new ArrayList(); 

  // Determine how many stitches we will be generating
  int numToGenerate = 0;
  if (numStitches == 0) numToGenerate = ((int) random(5)) + 1;
  else numToGenerate = numStitches;

  // Generate the stitches to fill the num with chance of repetition
  for (int i = 0; i < numToGenerate; i++) {
    // Generate a random first stitch
    if (i == 0) {
      int stRand = ((int) random(coreSts.length-4)) + 4;
      stitches.add(coreSts[stRand].copyStitch());
    } 
    else {
      // If there is a repeat chance, check it
      int repeatRand = ((int) random(100));
      if (repeatRand < chanceOfRepetition) {
        // If the 0-100 random number < number selected, repeat the last stitch
        Stitch lastStitch = (Stitch) stitches.get(i-1);
        stitches.add(coreSts[lastStitch.coreStID].copyStitch());
      } 
      else { // Else, find a completely random stitch
        int stRand = ((int) random(coreSts.length-4)) + 4;
        stitches.add(coreSts[stRand].copyStitch());
      }
    }
  }

  // Now we have x number of stitches with repetition
  // Apply symmetry, if any, and if we do, check if we need to input a chain.
  int symmetryRand = ((int) random(100));
  if (symmetryRand < chanceOfSymmetry) {
    symmetrical = true;
    int middleID = numToGenerate/2;
    // If we have an odd number of stitches...
    if (insertChainAtPeak && numToGenerate % 2 == 1) {
      // If chain @ peak is selected, replace the middle stitch with a chain
      stitches.set(middleID, coreSts[4].copyStitch());
    } 
    else if (insertChainAtPeak) {
      // If chain @ peak is selected, insert a stitch in the middle
      stitches.add(middleID, coreSts[4].copyStitch());
    }
    // Now repeat the left side on the right side in mirrored form.
    for (int j = 0; j < middleID; j++) {
      Stitch st = (Stitch) stitches.get(j);
      int reflectionID = stitches.size()-1-j;
      stitches.set(reflectionID, st.copyStitch());
    }
  } 
  else {
    symmetrical = false;
    // Do nothing! No Symmetry, no Symmetrical peak, the creation of stitches is done.
  }
  generateClusters();
}

void generateClusters() {
  clusters = new ArrayList();
  pieces = new ArrayList();

  // Determining if we're clustering
  int clusterRand = ((int) random(100));
  if (clusterRand < chanceToGroup) clustersOn = true;
  else clustersOn = false;

  if (clustersOn) {
    int numToCluster = ((int) random(6)) + 2; // Group 2-7 stitches
    if (symmetrical) {
      // Cluster around the symmetrical midpoint
      int middleID = stitches.size()/2;
      Cluster clust = null;
      // Odd number of stitches to cluster, make sure to include the peak
      int startIndex;
      if (stitches.size() % 2 == 0 && numToCluster % 2 == 0) startIndex = middleID - (numToCluster/2) -1;
      else startIndex = middleID - (numToCluster/2);
      for (int i = 0; i < stitches.size(); i++) {
        // any pre- or post-cluster, add them as they're own cluster
        if (i < startIndex || i > startIndex + numToCluster) {
          Stitch st = (Stitch) stitches.get(i);
          clust = new Cluster(st.copyStitch());
          clusters.add(clust);
          clust = null;
        } 
        else { // Else, add them to the working cluster
          // if the new cluster hasn't been started yet, add it
          if (clust == null) clust = new Cluster();
          Stitch st = (Stitch) stitches.get(i);
          clust.appendStitch(st.copyStitch());
        }
        if (i == startIndex + numToCluster) {
          // Finish off the cluster and do single from then on 
          clusters.add(clust);
          clust = null;
        }
      }
    } 
    else {
      Cluster clust = null;
      int clusterCt = 0;
      for (int i = 0; i < stitches.size(); i++) {
        if (clust == null) clust = new Cluster();
        Stitch st = (Stitch) stitches.get(i);
        clust.appendStitch(st.copyStitch());
        clusterCt++;
        if (i == stitches.size()-1 || clusterCt == numToCluster || clusterRand >= chanceToGroup) {
          clusters.add(clust);
          clust = null;
          clusterCt = 0;
        }
        clusterRand = ((int) random(100));
      }
    }
  } 
  else {
    // If clusters aren't turned on, dump them each into their own cluster and move on.
    for (int i = 0; i < stitches.size(); i++) {
      Stitch st = (Stitch) stitches.get(i);
      Cluster clust = new Cluster(st.copyStitch());
      clusters.add(clust);
    }
  }
  pieces.add(new PatternPiece(clusters));
  generateFillerStitches();
  generateAlternateCluster();
  generateRow();
}

void generateFillerStitches() {
  fillerPieces = new ArrayList();
  // If filler stitches is turned on...
  if (numFillerStitches > 0) {
    // Pick a random stitch that requires an anchor
    int stRand = ((int) random(coreSts.length-5)) + 5;
    Stitch st = coreSts[stRand].copyStitch();
    PatternPiece p = new PatternPiece();
    // And fill up a cluster with the specified number of that stitch
    for (int i = 0; i < numFillerStitches; i++) {
      Cluster clust = new Cluster(st.copyStitch());
      p.appendCluster(clust);
    }
    // Keep track of this piece separately for later use
    fillerPieces.add(p);
  }
}

void generateAlternateCluster() {
  PatternPiece reversePiece = new PatternPiece();
  if (horizontalReversal && pieces.size() > 0) {
    PatternPiece p = (PatternPiece) pieces.get(0);
    // For each cluster, starting with the last one
    for (int i = p.clusters.size() -1; i >= 0; i--) {
      Cluster clustOld = (Cluster) p.clusters.get(i);
      Cluster clustNew = new Cluster();
      // Reverse the stitches found in it!
      for (int j = clustOld.clustSts.size()-1; j >= 0; j--) {
        Stitch st = (Stitch)clustOld.clustSts.get(j);
        clustNew.appendStitch(st.copyStitch());
      }
      reversePiece.appendCluster(clustNew);
    }
    pieces.add(reversePiece);
  }
}

void generateRow() {
  rows = new ArrayList();
  Row r = new Row();
  PatternPiece p;
  int lengthOfRow = lengthRows;

  // Start with filler, end with filler
  if (numFillerStitches > 0) {
    for (int k = 0; k < fillerPieces.size(); k++) {
      p = (PatternPiece) fillerPieces.get(k);
      r.appendPiece(p);
    }
  }
  lengthOfRow--;

  // Fill in the rest of the pattern
  for (int i = 0; i < lengthOfRow; i++) {
    for (int j = 0; j < pieces.size(); j++) {
      p = (PatternPiece) pieces.get(j);
      r.appendPiece(p);
    }
    for (int k = 0; k < fillerPieces.size(); k++) {
      p = (PatternPiece) fillerPieces.get(k);
      r.appendPiece(p);
    }
  }
  rows.add(r);
}

